package eu.jonahbauer.tichu.common.model;

import eu.jonahbauer.tichu.common.model.exceptions.InvalidCombinationException;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Unmodifiable;

import java.util.List;
import java.util.Objects;

/**
 * A combination of cards.
 * @param type the type of the combination
 * @param cards the cards composing this combination
 */
@SuppressWarnings("unused")
public record Combination(
        @NotNull CombinationType type,
        @NotNull @Unmodifiable List<@NotNull Card> cards
) {
    /**
     * Creates a new combination.
     * @param type the type of the new combination
     * @param cards the cards composing the new combination
     * @throws InvalidCombinationException if the cards do not compose a valid combination of the given type
     * @throws NullPointerException if {@code type}, {@code cards} or any element of {@code cards} is {@code null}
     */
    public Combination(@NotNull CombinationType type, @NotNull List<@NotNull Card> cards) {
        this.type = Objects.requireNonNull(type, "type must not be null");
        this.cards = List.copyOf(Objects.requireNonNull(cards, "cards must not be null")); // List#copyOf performs null-checks
        if (!type.check(this.cards)) {
            throw new InvalidCombinationException(this.type, this.cards);
        }
    }

    /**
     * Returns the card at the specified position. This is equivalent to {@code cards().get(index)}
     * @param index index of the card to return
     * @return the card at the specified position
     * @see #cards()
     */
    public @NotNull Card card(int index) {
        return cards.get(index);
    }

    @Override
    public @NotNull String toString() {
        return type + cards.toString();
    }

    //<editor-fold desc="Named Constructors" defaultstate="collapsed">
    /**
     * Creates a new {@link CombinationType#SINGLE}-Combination.
     * @param card the card
     * @return a new combination
     * @see CombinationType#SINGLE
     */
    @Contract("_ -> new")
    public static @NotNull Combination single(@NotNull Card card) {
        return new Combination(CombinationType.SINGLE, List.of(card));
    }

    /**
     * Creates a new {@link CombinationType#PAIR}-Combination.
     * @param card0 the first card
     * @param card1 the second card
     * @return a new combination
     * @see CombinationType#PAIR
     */
    @Contract("_, _ -> new")
    public static @NotNull Combination pair(@NotNull Card card0, @NotNull Card card1) {
        return new Combination(CombinationType.PAIR, List.of(card0, card1));
    }

    /**
     * Creates a new {@link CombinationType#TRIPLE}-Combination.
     * @param card0 the first card
     * @param card1 the second card
     * @param card2 the third card
     * @return a new combination
     * @see CombinationType#TRIPLE
     */
    @Contract("_, _, _ -> new")
    public static @NotNull Combination triple(@NotNull Card card0, @NotNull Card card1, @NotNull Card card2) {
        return new Combination(CombinationType.TRIPLE, List.of(card0, card1, card2));
    }

    /**
     * Creates a new {@link CombinationType#FULL_HOUSE}-Combination.
     * @param tripleCard0 the first card of the triple
     * @param tripleCard1 the second card of the triple
     * @param tripleCard2 the third card of the triple
     * @param pairCard0 the first card of the pair
     * @param pairCard1 the second card of the pair
     * @return a new combination
     * @see CombinationType#FULL_HOUSE
     */
    @Contract("_, _, _, _, _ -> new")
    public static @NotNull Combination fullHouse(
            @NotNull Card tripleCard0, @NotNull Card tripleCard1, @NotNull Card tripleCard2,
            @NotNull Card pairCard0, @NotNull Card pairCard1
    ) {
        return new Combination(CombinationType.FULL_HOUSE, List.of(
                tripleCard0, tripleCard1, tripleCard2, pairCard0, pairCard1
        ));
    }

    /**
     * Creates a new {@link CombinationType#SEQUENCE}-Combination.
     * @param cards the sequence of cards (in correct order)
     * @return a new combination
     * @see CombinationType#SEQUENCE
     */
    @Contract("_ -> new")
    public static @NotNull Combination sequence(@NotNull Card @NotNull... cards) {
        return new Combination(CombinationType.SEQUENCE, List.of(cards));
    }

    /**
     * Creates a new {@link CombinationType#SEQUENCE}-Combination.
     * @param cards the sequence of cards (in correct order)
     * @return a new combination
     * @see CombinationType#SEQUENCE
     */
    @Contract("_ -> new")
    public static @NotNull Combination sequence(@NotNull List<@NotNull Card> cards) {
        return new Combination(CombinationType.SEQUENCE, cards);
    }

    /**
     * Creates a new {@link CombinationType#PAIR_SEQUENCE}-Combination.
     * @param cards the sequence of cards (in correct order)
     * @return a new combination
     * @see CombinationType#PAIR_SEQUENCE
     */
    @Contract("_ -> new")
    public static @NotNull Combination pairSequence(@NotNull Card @NotNull... cards) {
        return new Combination(CombinationType.PAIR_SEQUENCE, List.of(cards));
    }

    /**
     * Creates a new {@link CombinationType#PAIR_SEQUENCE}-Combination.
     * @param cards the sequence of cards (in correct order)
     * @return a new combination
     * @see CombinationType#PAIR_SEQUENCE
     */
    @Contract("_ -> new")
    public static @NotNull Combination pairSequence(@NotNull List<@NotNull Card> cards) {
        return new Combination(CombinationType.PAIR_SEQUENCE, cards);
    }

    /**
     * Creates a new {@link CombinationType#BOMB}-Combination.
     * @param card0 the first card
     * @param card1 the second card
     * @param card2 the third card
     * @param card3 the fourth card
     * @return a new combination
     * @see CombinationType#BOMB
     */
    @Contract("_, _, _, _ -> new")
    public static @NotNull Combination bomb(
            @NotNull Card card0, @NotNull Card card1, @NotNull Card card2, @NotNull Card card3
    ) {
        return new Combination(CombinationType.BOMB, List.of(card0, card1, card2, card3));
    }

    /**
     * Creates a new {@link CombinationType#BOMB_SEQUENCE}-Combination.
     * @param cards the sequence of cards (in correct order)
     * @return a new combination
     * @see CombinationType#BOMB_SEQUENCE
     */
    @Contract("_ -> new")
    public static @NotNull Combination bombSequence(@NotNull Card @NotNull... cards) {
        return new Combination(CombinationType.BOMB_SEQUENCE, List.of(cards));
    }

    /**
     * Creates a new {@link CombinationType#BOMB_SEQUENCE}-Combination.
     * @param cards the sequence of cards (in correct order)
     * @return a new combination
     * @see CombinationType#BOMB_SEQUENCE
     */
    @Contract("_ -> new")
    public static @NotNull Combination bombSequence(@NotNull List<@NotNull Card> cards) {
        return new Combination(CombinationType.BOMB_SEQUENCE, cards);
    }
    //</editor-fold>
}
