package eu.jonahbauer.tichu.common.model;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Range;

import java.util.*;

/**
 * A type of combination.
 */
public enum CombinationType {
    /**
     * A single card, {@linkplain Card#isNormal() normal} or {@linkplain Card#isSpecial() special}. This is the
     * correct {@code CombinationType} for playing the {@link Card#HOUND} or {@link Card#DRAGON}.
     */
    SINGLE {
        private static final Set<Card> SPECIAL_CARDS = EnumSet.of(Card.PHOENIX, Card.MAHJONG, Card.DRAGON, Card.HOUND);

        @Override
        public boolean check(@NotNull List<@NotNull Card> cards) {
            return checkBasics(cards, 1, SPECIAL_CARDS);
        }

        @Override
        public boolean isHigher(@NotNull Stack stack, @NotNull List<@NotNull Card> cards) {
            assert isCompatible(stack, cards);
            if (stack.isEmpty()) return true; // you can play any card onto an empty stack
            if (stack.size() == 1 && stack.top().card(0) == Card.HOUND) return false; // you cannot play any card onto a hound

            var card = cards.get(0);
            return switch (card) {
                case HOUND, MAHJONG -> false;
                case DRAGON -> true;
                case PHOENIX -> stack.top().card(0) != Card.DRAGON;
                default -> {
                    var top = stack.top().card(0);
                    if (top != Card.PHOENIX) {
                        yield card.getValue() > top.getValue();
                    } else if (stack.size() == 1) {
                        // every normal card has a value of at least 2, which is higher that the phoenix 1½
                        yield true;
                    } else {
                        yield card.getValue() > stack.top(-1).card(0).getValue();
                    }
                }
            };
        }
    },
    /**
     * A pair, i.e. two {@linkplain Card#isNormal() normal cards} of the same {@linkplain Card#getValue() value} or the
     * {@link Card#PHOENIX} and a normal card.
     */
    PAIR {
        private static final Set<Card> SPECIAL_CARDS = EnumSet.of(Card.PHOENIX);
        private static final int[] INDICES = new int[] {0, 1};

        @Override
        public boolean check(@NotNull List<@NotNull Card> cards) {
            return checkBasics(cards, 2, SPECIAL_CARDS)
                    && isSameValue(cards, INDICES);
        }

        @Override
        public boolean isHigher(@NotNull Stack stack, @NotNull List<@NotNull Card> cards) {
            assert isCompatible(stack, cards);
            return stack.isEmpty()
                   || getSameValue(cards, INDICES) > getSameValue(stack.top().cards(), INDICES);
        }
    },
    /**
     * A triple, i.e. three {@linkplain Card#isNormal() normal cards} of the same {@linkplain Card#getValue() value}
     * or the {@linkplain Card#PHOENIX} and two normal cards of the same value.
     */
    TRIPLE {
        private static final Set<Card> SPECIAL_CARDS = EnumSet.of(Card.PHOENIX);
        private static final int[] INDICES = new int[] {0, 1, 2};

        @Override
        public boolean check(@NotNull List<@NotNull Card> cards) {
            return checkBasics(cards, 3, SPECIAL_CARDS)
                    && isSameValue(cards, INDICES);
        }

        @Override
        public boolean isHigher(@NotNull Stack stack, @NotNull List<@NotNull Card> cards) {
            assert isCompatible(stack, cards);
            return stack.isEmpty()
                    || getSameValue(cards, INDICES) > getSameValue(stack.top().cards(), INDICES);
        }
    },
    /**
     * A full house, i.e. five cards of which three form a {@link #TRIPLE} and the other two form a {@link #PAIR}.
     * @apiNote A list of cards representing a full house must have the triple at the first three indices and the pair
     * at the last two. This prevents any ambiguity, e.g. when the full house consists of two pairs and the phoenix.
     */
    FULL_HOUSE {
        private static final Set<Card> SPECIAL_CARDS = EnumSet.of(Card.PHOENIX);

        private static final int[] INDICES_TRIPLE = new int[] {0, 1, 2};
        private static final int[] INDICES_PAIR = new int[] {3, 4};

        @Override
        public boolean check(@NotNull List<@NotNull Card> cards) {
            return checkBasics(cards, 5, SPECIAL_CARDS)
                    && isSameValue(cards, INDICES_TRIPLE)
                    && isSameValue(cards, INDICES_PAIR);
        }

        @Override
        public boolean isHigher(@NotNull Stack stack, @NotNull List<@NotNull Card> cards) {
            assert isCompatible(stack, cards);
            return stack.isEmpty()
                   || getSameValue(cards, INDICES_TRIPLE) > getSameValue(stack.top().cards(), INDICES_TRIPLE);
        }
    },
    /**
     * A sequence, i.e. at least five cards with increasing values. The first card in the sequence may be a
     * {@link Card#MAHJONG} and one of the cards may be the {@link Card#PHOENIX}.
     * @apiNote A list of cards representing a sequence must be sorted by ascending value. This allows to distinguish
     * between a sequence starting with the phoenix and a sequence ending with the phoenix.
     */
    SEQUENCE {
        private static final Set<Card> SPECIAL_CARDS = EnumSet.of(Card.PHOENIX, Card.MAHJONG);

        private int getStartValue(@NotNull List<@NotNull Card> cards) {
            return cards.get(0) == Card.PHOENIX
                    ? cards.get(1).getValue() - 1
                    : cards.get(0).getValue();
        }

        @Override
        public boolean check(@NotNull List<@NotNull Card> cards) {
            if (!checkBasics(cards, null, SPECIAL_CARDS)) return false;
            if (cards.size() < 5) return false;

            int start = getStartValue(cards);

            for (int i = 0; i < cards.size(); i++) {
                var card = cards.get(i);
                if (card == Card.PHOENIX) {
                    if (start + i > 14) {
                        return false;
                    }
                } else if (card.getValue() != start + i) {
                    return false;
                }
            }

            return true;
        }

        @Override
        public boolean isCompatible(@NotNull Stack stack, @NotNull List<@NotNull Card> cards) {
            assert check(cards);
            return stack.isEmpty()
                   || stack.top().type() == this && stack.top().cards().size() == cards.size();
        }

        @Override
        public boolean isHigher(@NotNull Stack stack, @NotNull List<@NotNull Card> cards) {
            assert isCompatible(stack, cards);
            return stack.isEmpty()
                   || getStartValue(cards) > getStartValue(stack.top().cards());
        }
    },
    /**
     * A sequence of pairs, i.e. at least two pairs with increasing values.
     * @apiNote Although not strictly necessary, for consistency with {@link #SEQUENCE} a list of cards representing
     * a sequence of pairs must be sorted by ascending value.
     */
    PAIR_SEQUENCE {
        private static final Set<Card> SPECIAL_CARDS = EnumSet.of(Card.PHOENIX);

        private int getStartValue(@NotNull List<@NotNull Card> cards) {
            return cards.get(0) == Card.PHOENIX
                    ? cards.get(1).getValue()
                    : cards.get(0).getValue();
        }

        @Override
        public boolean check(@NotNull List<@NotNull Card> cards) {
            if (!checkBasics(cards, null, SPECIAL_CARDS)) return false;
            if (cards.size() < 4 || cards.size() % 2 == 1) return false;

            int start = getStartValue(cards);

            for (int i = 0; i < cards.size(); i++) {
                var card = cards.get(i);

                if (card != Card.PHOENIX && card.getValue() != start + i / 2) {
                    return false;
                }
            }

            return true;
        }

        @Override
        public boolean isCompatible(@NotNull Stack stack, @NotNull List<@NotNull Card> cards) {
            assert check(cards);
            return stack.isEmpty()
                    || stack.top().type() == this && stack.top().cards().size() == cards.size();
        }

        @Override
        public boolean isHigher(@NotNull Stack stack, @NotNull List<@NotNull Card> cards) {
            assert isCompatible(stack, cards);
            return stack.isEmpty()
                    || getStartValue(cards) > getStartValue(stack.top().cards());
        }
    },
    /**
     * A bomb, i.e. four {@linkplain Card#isNormal() normal cards} of the same {@linkplain Card#getValue() value}.
     */
    BOMB {
        private static final int[] INDICES = new int[] {0, 1, 2, 3};
        private static final Set<Card> SPECIAL_CARDS = Collections.emptySet();
        private static final Combination COMBINATION_HOUND = new Combination(SINGLE, List.of(Card.HOUND));

        @Override
        public boolean check(@NotNull List<@NotNull Card> cards) {
            return checkBasics(cards, 4, SPECIAL_CARDS)
                    && isSameValue(cards, INDICES);
        }

        @Override
        public boolean isCompatible(@NotNull Stack stack, @NotNull List<@NotNull Card> cards) {
            assert check(cards);
            return stack.isEmpty() || !COMBINATION_HOUND.equals(stack.top());
        }

        @Override
        public boolean isHigher(@NotNull Stack stack, @NotNull List<@NotNull Card> cards) {
            assert isCompatible(stack, cards);
            return stack.isEmpty()
                    || stack.top().type() != BOMB && stack.top().type() != BOMB_SEQUENCE
                    || stack.top().type() == BOMB && getSameValue(cards, INDICES) > getSameValue(stack.top().cards(), INDICES);
        }
    },
    /**
     * A bomb sequence, i.e. a {@linkplain Card#getSuit() monochrome} {@link #SEQUENCE} that does not contain any
     * {@linkplain Card#isSpecial() special cards}.
     * @apiNote Although not strictly necessary, for consistency with {@link #SEQUENCE} a list of cards representing a
     * bomb sequence must be sorted by ascending value.
     * @apiNote
     */
    BOMB_SEQUENCE {
        private static final Set<Card> SPECIAL_CARDS = Collections.emptySet();
        private static final Combination COMBINATION_HOUND = new Combination(SINGLE, List.of(Card.HOUND));

        private int getStartValue(@NotNull List<@NotNull Card> cards) {
            return cards.get(0).getValue();
        }

        @Override
        public boolean check(@NotNull List<@NotNull Card> cards) {
            if (!checkBasics(cards, null, SPECIAL_CARDS)) return false;
            if (cards.size() < 5) return false;
            if (cards.stream().map(Card::getSuit).distinct().count() != 1) return false;

            int start = getStartValue(cards);
            for (int i = 1; i < cards.size(); i++) {
                var card = cards.get(i);
                if (card.getValue() != start + i) {
                    return false;
                }
            }

            return true;
        }

        @Override
        public boolean isCompatible(@NotNull Stack stack, @NotNull List<@NotNull Card> cards) {
            assert check(cards);
            return stack.isEmpty() || !COMBINATION_HOUND.equals(stack.top());
        }

        @Override
        public boolean isHigher(@NotNull Stack stack, @NotNull List<@NotNull Card> cards) {
            assert isCompatible(stack, cards);
            return stack.isEmpty()
                   || stack.top().type() != BOMB_SEQUENCE
                   || cards.size() > stack.top().cards().size()
                   || getStartValue(cards) > getStartValue(stack.top().cards());
        }
    };

    /**
     * Checks whether the given cards form a valid combination of this type. See the documentation of a specific
     * {@code CombinationType} for more information on what cards form a combination of that particular type.
     * @param cards a list of cards
     * @return {@code true} if the cards form a valid combination of this type.
     */
    public abstract boolean check(@NotNull List<@NotNull Card> cards);

    /**
     * <p>Checks whether the {@code cards} are <i>compatible</i> with the {@code stack}, i.e. the {@code cards} can
     * theoretically be played on top of the {@code stack}. For example a {@link #SEQUENCE} is compatible with a given
     * stack only if the stack is empty, or it consists of sequences of the same length.
     * <p>However, this method does not check whether the {@code cards} are actually higher than the top of the
     * {@code stack}. If the {@code cards} are not a {@linkplain #check(List) valid} combination of this type, the
     * behaviour of this method is undefined and an exception may be thrown.
     * @param stack the current stack of cards
     * @param cards the cards to be played
     * @return {@code true} iff the {@code cards} are compatible with the {@code stack}
     */
    public boolean isCompatible(@NotNull Stack stack, @NotNull List<@NotNull Card> cards) {
        assert check(cards);
        return stack.isEmpty() || stack.top().type() == this;
    }

    /**
     * <p>Checks whether the {@code cards} are higher than the combination currently on top of the {@code stack}.
     * <p>If the {@code cards} are not {@linkplain #isCompatible(Stack, List) compatible} with the {@code stack},
     * the behaviour of this method is undefined and an exception may be thrown.
     * @param stack the current stack of cards
     * @param cards the cards to be played
     * @return {@code true} iff the {@code cards} can be played on top of the {@code stack}
     */
    public abstract boolean isHigher(@NotNull Stack stack, @NotNull List<@NotNull Card> cards);

    //<editor-fold desc="Utility Methods" defaultstate="collapsed">
    /**
     * Performs basic checks on a list of cards, i.e. this method checks
     * <ul>
     *     <li>that the list does not contain any duplicate cards</li>
     *     <li>when {@code count != null}, that the list is of size {@code count}</li>
     *     <li>
     *         that the list does not contain any {@linkplain Card#isSpecial() special card} other than those in
     *         {@code special}
     *     </li>
     * </ul>
     * @param cards the list of cards
     * @param count the expected number of cards, or {@code null} when the number of cards should not be checked
     * @param special the set of allowed special cards or {@code null}, when all special cards are allowed.
     */
    private static boolean checkBasics(@NotNull List<@NotNull Card> cards, Integer count, @NotNull Set<@NotNull Card> special) {
        if (count != null && cards.size() != count) return false;
        if (containsDuplicates(cards)) return false;

        for (Card card : cards) {
            if (card.isSpecial() && !special.contains(card)) {
                return false;
            }
        }

        return true;
    }

    /**
     * Checks whether the list contains any duplicate cards, i.e. there exists {@code a} and {@code b} such that
     * <pre><code>cards.contains(a) && cards.contains(b) && a.equals(b)</code></pre>
     * @param cards a list of cards
     * @return {@code true} iff the list contains duplicates.
     */
    private static boolean containsDuplicates(@NotNull List<@NotNull Card> cards) {
        if (cards.size() <= 1) return false;
        var set = EnumSet.noneOf(Card.class);
        for (Card card : cards) {
           if (!set.add(card)) return true;
        }
        return false;
    }

    /**
     * Checks that the cards at the specified indices all have the same value.
     * {@linkplain Card#isSpecial() Special cards} are ignored.
     * @param cards a list of cards
     * @param indices an array of indices of cards in the list
     * @return {@code true} iff all of the specified cards have the same value
     */
    private static boolean isSameValue(@NotNull List<@NotNull Card> cards, int @NotNull... indices) {
        Integer value = null;

        for (int index : indices) {
            Card card = cards.get(index);
            if (card.isSpecial()) continue;

            if (value == null) {
                value = card.getValue();
            } else if (!value.equals(card.getValue())) {
                return false;
            }
        }

        return true;
    }

    /**
     * Assuming that all the {@linkplain Card#isNormal() normal cards} at the specified indices have the same value
     * returns that value. The return value is unspecified if the assumption is violated.
     * @param cards a list of cards
     * @param indices an array of indices of cards in the list
     * @see #isSameValue(List, int...)
     * @throws NoSuchElementException if there is no normal card among the specified indices
     */
    private static @Range(from = 2, to = 14) int getSameValue(@NotNull List<@NotNull Card> cards, int @NotNull... indices) {
        for (int index : indices) {
            Card card = cards.get(index);
            if (card.isSpecial()) continue;

            return card.getValue();
        }
        throw new NoSuchElementException();
    }
    //</editor-fold>
}
